// ------------------------------------------------------------------------ //
//			       LZW Compression				    //
//									    //
//			    Author: Jack Thornton			    //
//			       Level Five, Inc. 			    //
//		   952 Postal Way, Ste. 3, Vista, CA 92083		    //
//	 Ph: (619) 631-5855  Fax: (619) 631-5858  email: jack@cts.com	    //
// ------------------------------------------------------------------------ //

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>
#include "resource.h"
#include "lzwcomp.h"

// ***************************************************** Constant Declarations
// Bit-size related constants
#define MAX_BITS	12
#define MAX_CODE	((1 << MAX_BITS) - 1)
#define TABLE_SIZE	4919  /* should be prime and 20% larger than MAX_CODE */

// Special codes
#define END_OF_STREAM	256
#define BUMP_CODE	257
#define FLUSH_CODE	258
#define FIRST_CODE	259
#define UNUSED		((CODE)(-1))

// ******************************************** Structure and Type Definitions
typedef unsigned short CODE;	// allows for maximum 16-bit codes

// structure of individual entry in hash table
typedef struct dictionary {
    CODE codeValue;	// index (code value) of this dictionary entry
    CODE parentCode;	// index of "parent code" (prefix phrase)
    BYTE character;	// character added to make current phrase
} DictionaryEntry;

/* This is the structure of the output pseudo-file */
typedef struct {
    BYTE *data;
    BYTE mask;
    CODE rack;
    DWORD length;
} BIT_STREAM;

// **************************************************************** Prototypes
CODE HashChildNode(CODE parentCode, BYTE child_character);
void InitializeDictionary(void);
DWORD CompressFile(BYTE *, BYTE *, DWORD);
void OutputBits(BIT_STREAM *, CODE);

// external, ASM-hosted functions
extern "C" BOOL ExpandFile(BYTE *, BYTE *, DWORD, DWORD);

// macro-ized functions
#define GetByte(x) (*x++)
#define PutByte(x, y) (*y++ = x)

// ****************************************************** Variables and Tables
DictionaryEntry *dict;		    // phrase dictionary
CODE nextCode;			    // next code to be added to dictionary
DWORD codeBitLength;		    // current size of code, in bits
CODE nextBumpCode;		    // code that triggers increase in bit size
int curError;			    // current error
LPSTR errorStrings[] = {
    "No error",
    "Internal error of unknown origin (i.e. a bug)",
    "Bad parameters passed (NULL pointer or zero length)",
    "Out of memory",
    "Encoded data larger than original",
    "Corrupted data: decoder input data overrun",
    "Corrupted data: decoder output data overrun"
};

// ********************************************************* Code: Actual work
LPSTR
getLastError(void)
{
    // clip error code
    if (curError > MAX_ERR)
	curError = ERR_ASSERTION;

    // return message relating to error code
    return errorStrings[curError];
}

// Notes about the implementation:
// This implementation takes advantage of some of the refinements to the LZW
// algorithm seen in lots of the compression programs out there, such as
// SQUEEZE and the Unix COMPRESS programs. These improvements are:
// -  Variable bit-length encoding (starts at 9 bits)
// -  Flushes dictionary when full to avoid degredation of compression
// I didn't bother having the compressor do compression rate statistics to
// determine when to flush the dictionary - I just flush it when it gets full.
// An improvement to make sometime is to track the compression efficiency, and
// only flush the dictionary when the rate starts to deteriorate.
//

// InitializeDictionary
//   This routine is used to initialize the dictionary when the compressor
// initializes and when it flushes a full dictionary.
// --------------------------------------------------------------------------
void
InitializeDictionary()
{
DWORD i;

    for (i = 0 ; i < TABLE_SIZE; i++)
	dict[i].codeValue = UNUSED;
    nextCode = FIRST_CODE;
    codeBitLength = 9;
    nextBumpCode = 511;
}

// CompressFile
//   The compressor is short and simple.  It reads in new symbols one
// at a time from the input file.  It then checks to see if the
// combination of the current symbol and the current code are already
// defined in the dictionary.  If they are not, they are added to the
// dictionary, and we start over with a new one symbol code.  If they
// are, the code for the combination of the code and character becomes
// our new code.  Note that in this enhanced version of LZW, the
// encoder needs to check the codes for boundary conditions.
// --------------------------------------------------------------------------
DWORD
CompressFile(BYTE *output, BYTE *input, DWORD length)
{
BYTE character; 	    // current character (last one read from input)
CODE stringCode;	    // index of currently matching phrase
DictionaryEntry *ep;	    // dereference pointer
BIT_STREAM out;
DWORD sourceCount;

    // parameter test
    if (!input || !output || !length) {
	curError = ERR_BADPARAMS;
	return 0;
    }

    // allocate a dictionary
    dict = (DictionaryEntry *) malloc(sizeof(DictionaryEntry) * TABLE_SIZE);
    if (!dict) {
	curError = ERR_MEMORY;
	return 0;
    }

    // Initialize the output bit stream
    out.data = output;
    out.rack = 0;
    out.mask = 0x80;
    out.length = 0;

    // initialize the phrase dictionary hash-table
    InitializeDictionary();

    // get initial byte to code
    sourceCount = length;
    stringCode = (CODE) GetByte(input);
    if (!sourceCount--)
	stringCode = END_OF_STREAM;

    // main encode loop (while there is still stuff to encode!)
    while (sourceCount--) {
	character = GetByte(input);

	// look up entry in dictionary to see if phrase so far matches
	ep = &dict[HashChildNode(stringCode, character)];
	if (ep->codeValue != UNUSED)
	    stringCode = ep->codeValue;
	else {
	    // no - need to make a new code
	    ep->codeValue = nextCode++;
	    ep->parentCode = stringCode;
	    ep->character = character;

	    // put new code to output stream
	    OutputBits(&out, stringCode);

	    // reinitialize code to start at the last read character
	    stringCode = (CODE) character;

	    // check to see if time to flush dictionary
	    if (nextCode > MAX_CODE) {
		OutputBits(&out, FLUSH_CODE);
		InitializeDictionary();

	    // or check to see if time to increase size of codes
	    } else if (nextCode > nextBumpCode) {
		OutputBits(&out, BUMP_CODE);
		codeBitLength++;
		//assert(codeBitLength <= MAX_BITS);
		nextBumpCode = (nextBumpCode << 1) | 1;
	    }
	}

	// check for overrun
	if (out.length >= length) {
	    curError = ERR_LARGER;
	    free(dict);
	    return 0;
	}
    }

    // flush last code, as well as the end-of-stream marker
    OutputBits(&out, stringCode);
    OutputBits(&out, END_OF_STREAM);

    // ensure bitrack is flushed
    if (out.mask != 0x80) {
	codeBitLength = 7;
	OutputBits(&out, 0);
    }

    // return length of compressed data stream
    free(dict);
    return out.length;
}

// HashChildNode
//   This hashing routine is responsible for finding the table location
// for a string/character combination.	The table index is created
// by using an exclusive OR combination of the prefix and character.
// This code also has to check for collisions, and handles them by
// jumping around in the table.
// --------------------------------------------------------------------------
CODE
HashChildNode(CODE parentCode, BYTE childCharacter)
{
CODE index;
CODE offset;
DictionaryEntry *ep;

    // get initial guess at location in table (hash value)
    index = ((CODE)childCharacter << (MAX_BITS - 8)) ^ parentCode;
    offset = (!index) ? 1 : TABLE_SIZE - index;
    //assert(index < TABLE_SIZE);
    //assert(offset < TABLE_SIZE);

    // lookup loop
    for ( ; ; ) {
	// dereference pointer into dictionary
	ep = &dict[index];

	// did we hit either a match or an unused code?
	if ((ep->codeValue == UNUSED) ||
	   ((ep->parentCode == parentCode) &&
	   (ep->character == childCharacter)))
		return index;	// yes - return index to caller

	// nope - jump to another location in the table
	if (index >= offset)
	    index -= offset;
	else
	    index += TABLE_SIZE - offset;
	//assert(index < TABLE_SIZE);
    }
}

void
OutputBits(BIT_STREAM *bits, CODE code)
{
CODE mask;

    //assert(bits);
    mask = 1 << (codeBitLength - 1);
    while (mask != 0) {
	if (mask & code)
	    bits->rack |= bits->mask;
	bits->mask >>= 1;
	if (bits->mask == 0) {
	    PutByte((BYTE)bits->rack, bits->data);
		bits->rack = 0;
		bits->mask = 0x80;
		bits->length++;
	}
	mask >>= 1;
    }
}

