//
// -----------------
// Splash Screen DLL
// -----------------
//	Provides a method for the GameFilm engine and CD-Autorun programs to
//	share bitmap resources. To save space, the bitmaps are stored as
//	compressed resources, with a special DLL function provided to
//	decompress the data, set up a DIB from the data (as a DIBSECTION)
//	and return an HBITMAP for that DIB.
//
//	Author: Jack Thornton
//	for CetaSoft / Level Five, Inc.
//	952 Postal Way, Suite 3
//	Vista, CA 92083
//	(619) 631-5855
//
//	Client: Atari Corporation
//	Project: GameFilm Engine
//	Date: 2/24/96
// --------------------------------------------------------------------------

// ******************************************************************* Headers
#include <windows.h>

// ****************************************************** Constants and Macros
#define STRID_NUMBITMAPS    1

// ****************************************************** Structures and Types
typedef struct {
    int refCount;		// reference counted data
    void *pBits;		// points to bitmap data (in a DIB format)
    BITMAPINFO *pHeader;	// points to a DIB header
    RECT destRect;		// intended destination rectangle of bitmap
    void *fileBuffer;		// if non-NULL, points to buffer to be freed
} BitmapResource;

typedef struct {
    DWORD originalSize;
    DWORD compSize;
    WORD x;
    WORD y;
    WORD width;
    WORD height;
} CompFileHeader;

// **************************************************************** Prototypes
// our exports
extern "C" __declspec(dllexport) BOOL APIENTRY FetchBitmap(UINT, LPBITMAPINFO *,
    LPVOID *, RECT *);
extern "C" __declspec(dllexport) void APIENTRY DiscardBitmap(UINT);

// external ASM provided functions for our benefit
extern "C" BOOL ExpandFile(BYTE *dest, BYTE *src, DWORD destLen, DWORD srcLen);

// other stuff
BOOL APIENTRY DllMain(HANDLE, DWORD, LPVOID);
BOOL initializeProcess(void);
BOOL cleanupProcess(void);

// ***************************************************************** Variables
HINSTANCE hModule;	    // handle to our module (in this process)
unsigned int numBitmaps;    // number of bitmaps in the resource
BitmapResource *resTable;   // dynamically allocated loaded resource table

// ********************************************************************** Code

BOOL APIENTRY
DllMain(HANDLE hMod, DWORD reason, LPVOID reserved)
{
    // why are we being called?
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	    // copy instance handle to process workspace
	    hModule = hMod;

	    // call initialization routine
	    return initializeProcess();

	case DLL_PROCESS_DETACH:
	    // call cleanup routine for this module
	    return cleanupProcess();

	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	default:
	    return TRUE;
    }
}

BOOL
initializeProcess()
{
char stringBuf[8];

    // load the string that tells us how many bitmaps are in the resource
    LoadString(hModule, STRID_NUMBITMAPS, stringBuf, 8);

    // turn the string into an integer
    numBitmaps = atoi(stringBuf);
    if (!numBitmaps)
	return FALSE;

    // allocate a structure table for the bitmaps
    resTable = (BitmapResource *) malloc(numBitmaps * sizeof(BitmapResource));
    if (!resTable)
	return FALSE;

    // clear the structure entries to NULL/ZERO/ZILCH
    memset(resTable, 0, numBitmaps * sizeof(BitmapResource));

    // okay, shop's open for business!
    return TRUE;
}

BOOL
cleanupProcess(void)
{
    // free the allocated table
    if (resTable)
	free(resTable);
    numBitmaps = 0;
    return TRUE;
}

__declspec(dllexport)
BOOL APIENTRY
FetchBitmap(UINT index, LPBITMAPINFO *pHeader, LPVOID *pBits, RECT *pDestRect)
{
HRSRC hResource;	// handle of resource in file
HGLOBAL hOrigData;	// handle of loaded resource
BYTE *pOrigData;	// points to loaded resource in system memory
void *pBMPFile; 	// points to raw BMP file data
BYTE *decompBuffer;	// decompression buffer
#define pResHeader     ((CompFileHeader *)pOrigData)
#define pData	       (&pOrigData[sizeof(CompFileHeader)])
#define pBMPFileHeader ((BITMAPFILEHEADER *)pBMPFile)

    // check validity of index
    if (index >= numBitmaps)
	return FALSE;

    // is this one already loaded?
    if (!resTable[index].refCount) {
	// nope: load the resource in question and get a pointer to it
	hResource = FindResource(hModule, (LPCTSTR) index + 1, "CETASOFT");
	if (!hResource)
	    return FALSE;
	hOrigData = LoadResource(hModule, hResource);
	if (!hOrigData)
	    return FALSE;
	pOrigData = (BYTE *) LockResource(hOrigData);
	if (!pOrigData)
	    return FALSE;

	// do we need to decompress this information or not?
	if (pResHeader->originalSize != pResHeader->compSize) {
	    // allocate a decompression buffer
	    decompBuffer = (BYTE *) malloc(pResHeader->originalSize);
	    if (!decompBuffer)
		return FALSE;

	    // decompress the resource into the buffer
	    if (!ExpandFile(decompBuffer, pData, pResHeader->originalSize,
	      pResHeader->compSize)) {
		free(decompBuffer);
		return FALSE;
	    }

	    // raw BMP file is contained in the decompression buffer
	    pBMPFile = (void *) decompBuffer;
	} else {
	    // signal that we have no temporary buffer to free
	    decompBuffer = NULL;

	    // raw BMP file is contained in original resource
	    pBMPFile = (void *) pData;
	}

	// allocate and initialize the structure for this indexed bitmap
	resTable[index].pBits = (void *)
	    ((DWORD)pBMPFile + pBMPFileHeader->bfOffBits);
	resTable[index].pHeader = (BITMAPINFO *)
	    ((DWORD)pBMPFile + sizeof(BITMAPFILEHEADER));
	resTable[index].destRect.left = pResHeader->x;
	resTable[index].destRect.top = pResHeader->y;
	resTable[index].destRect.right = pResHeader->x + pResHeader->width;
	resTable[index].destRect.bottom = pResHeader->y + pResHeader->height;
	resTable[index].fileBuffer = decompBuffer;
    }

    // pass back the desired information
    *pHeader = resTable[index].pHeader;
    *pBits = resTable[index].pBits;

    // copy destination rectangle to clients RECT structure
    *pDestRect = resTable[index].destRect;

    // increment our reference count
    resTable[index].refCount++;

    // bye-bye!
    return TRUE;
}

__declspec(dllexport)
void APIENTRY
DiscardBitmap(UINT index)
{
    // validate index
    if (index < numBitmaps) {
	// is it in use?
	if (resTable[index].refCount) {
	    // decrement the reference count and see if it is zero
	    if (--resTable[index].refCount == 0) {
		// discard allocated memory (if any)
		if (resTable[index].fileBuffer)
		    free(resTable[index].fileBuffer);

		// clear the entry
		memset(&resTable[index], 0, sizeof(BitmapResource));
	    }
	}
    }
}

