#ifndef __CPNODES_H
#define __CPNODES_H
/*
  File/memory layout for cineplay playback structures.
  On disk, replace each pointer with absolute file offset
  of the data.
  The file should start with its file size, a DWORD. 
  The CPNode for the first clip should be stored just after
  the file size. It isn't important how the other nodes are
  found, as long as the links can be followed.
  Unused offsets/pointers should be zero.

  The first CPNode should be the CONTROL/TYPE=TERMINAL.

  The database should be built in parallel with construction
  of the file that contains all the clips. In this way, the
  clip numbers in the database will synchronize with the clip
  numbers in the clip file.
*/

#pragma pack(push, 1)

struct CPTrack;
typedef int ClipSpec;
typedef int VOCSpec;
enum TypeSpec
{ CP_SHOT=0
, CP_LOOP=1
, CP_LEADIN=2
, CP_LEADOUT=3
};
//// New: Nodes can be controls instead of clips.
enum CPControlSpec 
{ CPC_SWITCH         // Always follow best match of N links.
, CPC_COUNT          // Commutate among N links.
, CPC_BURN           // Follow best match, but only one follow per link.
, CPC_CHANGEDISK     // Disk change must occur between backlink and link.
, CPC_RELAY          // Use visit count of relayCoil node to choose link.
, CPC_TERMINAL       // Either start or end of game.
, CPC_CENSOR         // Allow prevention of paths based on Sex/Violence.
};
enum CPNodeSpec 
{ CPN_FILM
, CPN_CONTROL 
};
//// Film Nodes now have only two rules.
enum RuleSpec
{ CP_NORULE=0 // WAS CP_NONE, conflicted with GetDeviceCaps.
, CP_HYPER=1
};
//// Psycho numbers MAY be WORDS.
struct Psych
{
  WORD data[5];
  Psych& operator+=(const Psych& other)
  {
    data[0]+=other.data[0];
    data[1]+=other.data[1];
    data[2]+=other.data[2];
    data[3]+=other.data[3];
    data[4]+=other.data[4];
    return *this;
  }
  unsigned ManhattanMetric(const Psych& other) const
  {
    // assumption: long is sufficient to handle
    // subtraction of unsigned 16-bit ints without overflow.
    // assumption: the WORD type is unsigned 16-bit ints.
    unsigned r = 0;
    for (int i=0; i<5; i++)
    {
      long t = long(data[i]) - long(other.data[i]);
      if (t<0) t=-t;
      if (t>65535) t=65535;
      r += unsigned(t);
    }
    return r;
  }
};
//// Burn and Switch controls need psychometric data.
struct BurnSwitchData
{
  int links; // number of non-default links.
  Psych* pLinks; // psych value for each non-default link.
};
//// HYPER now includes a useful prompt.
struct HyperData
{
  WORD startFrame;
  WORD endFrame;
  WORD buttonHits;
  char* prompt; // NEW
};
//// Track and Segment remain the same.
struct CPSegment;
struct CPTrack
{
  DWORD segments;
  CPSegment* pSegments;
};
struct CPSegment
{
  WORD firstFrame;
  WORD lastFrame;
  VOCSpec voiceOver;
  char* prompt;
  Psych psych;
  WORD completion;
};
//// Data for a node when it is a FILM node.
struct CPFilm
{
  ClipSpec clip;
  TypeSpec type;
  RuleSpec rule;
  HyperData data;
  unsigned visits;
  int viewed; // viewed flag may be true when visits==0.
  int tracks;
  CPTrack* pTracks;
  // links are in enclosing CPNode
};
//// Data to complete Control spec, depending on control.
// Most Controls just need the link data to work.
struct CPNode;
union CPControlData
{
  unsigned newDisk;  // FOR CHANGEDISK
  CPNode* relayCoil; // FOR RELAY
  BurnSwitchData burn;     // FOR BURN or SWITCH
};
struct CPControl
{
  CPControlSpec controlType;
  CPControlData data;
  unsigned visitBits;
};
struct CPNode
{
  CPNodeSpec nodeType;
  union
  {
    CPFilm film;
    CPControl control;
  } data;
  int links;
  CPNode** pLinks;
  int backLinks;
  CPNode** pBackLinks;
};

#pragma pack(pop)

#endif