/********************************************************************\

	Film Class Module
	by Rudy Mathieu

\********************************************************************/

#include <assert.h>
#include "Film.h"
#include "dbgtst.h"

static long FadeOutOffset = 0;	// Offset from end of clip

// This is an easier way to callback resource manager
#define RMCALLBACK(x) {if (rmCallback) (*rmCallback)(this,rmParam,(x));}

int	Film::AddCallback(DWORD frame, FilmCallback fn, void *user)
{
  if (trigger)
  {
    EventStr* e = new EventStr;
    e->frame = frame;
    e->fn = fn;
    e->user = user;
    return trigger->addptr(e);
  }
  else
  {
    assert(!"Adding to null triggers");
    return 0;
  }
}

int	Film::RemoveCallback(int h)
{
  assert(trigger && "removing from null trigger");
  return trigger->remove(h);
}

void Film::Close(void)
{
  reader.CloseBuffer();

  Init();
}  

BOOL __fastcall Film::Open(int id)
{
  reader.Link(id);
  if (reader.CheckBuffer())
  {
    FRSRC_FILEENTRY *entry = reader.GetFileInformation();
    if (entry)
      DebugString("Opening Film %s,ID %i, Buffer No. %i \n",entry->filename,entry->id,id,NULL);

    return TRUE;
  }
  return FALSE;
}

// 	returns		
//  0	Ok
//	-1	Error
//	-2	Header data isn't ready yet
int Film::SetupFilm(void)
{
  int bytes = reader.ReallyReadHeader(&header);
  if (bytes != sizeof(header))
  {
    Close();
    return -2;
  }

  if (CheckSoundFormat()==FALSE)
  {
    Close();
    return -1;
  }

  if (header.fileSize < (unsigned long)(reader.GetBufferSize())) //.buffer->Size())) 
    statusALL_WILL_FIT=TRUE;	// Fits entirely in memory

  if (ravout->Display()->FormatDifferent(&header.bitmapInfo))
    statusVIDEODIFFERENT=TRUE;	// Needs to change display format

  return 0;
}

BOOL Film::Play(LONG loopcount, FADETYPE type, int incr)
{
  int err = SetupFilm();
  assert(err == 0); // This should not fail!!!

  if (header.totalframes==0)
  {
    // Done playing callback
    RMCALLBACK(FILMDONEPLAYING);
    Close();
    return TRUE;
  }

  FRSRC_FILEENTRY *entry = reader.GetFileInformation();
  if (entry)
    DebugString("Playing from file (Name-%s)...\n",entry->filename,NULL);
  else
    DebugString("Playing from buffer\n",NULL);


  loop = loopcount;
  loopCount = 0;
  clipfadetype = type;
  clipfadeinc = incr;

  statusPLAYING=TRUE;
  if (statusVIDEODIFFERENT==TRUE)
  {
    ravout->Display()->Setup(&header.bitmapInfo); /* Setup Video Info */
    statusVIDEODIFFERENT=FALSE;
  }

  Fade();

  return TRUE;
}

BOOL Film::Check(void)
{
  if (!Playing() || Paused())
    return FALSE;

  if (!statusSAMPLE_READY)
    if (ReadNextFrame()==0)
    statusSAMPLE_READY = TRUE;

  if (statusSAMPLE_READY)
    if (!ravout->SampleReady())
    if (ravout->Load(sample)==TRUE)
    statusSAMPLE_READY=FALSE;

  return TRUE;
}

int Film::ReadNextFrame(void)
{
  long rv = reader.ReadSample(sample);
  //assert(rv != SAMPLE_READ_ERROR);
  if (rv == SAMPLE_READ_ERROR)
  {
    DebugString("SAMPLE_READ_ERROR\n");
    RMCALLBACK(FILMDONEPLAYING);
    Close();
    return BUFFER_EMPTY;
  }

  if (rv==BUFFER_BUSY)
  {
    return rv;// Data not ready yet!!!
  }

  if (rv==BUFFER_EMPTY)
  {
    // Are we done yet?
    // Check looping()
    if (Looping()==FALSE)
    {
      // Done playing callback
      RMCALLBACK(FILMDONEPLAYING);

      Close(); // Close it up. We are done.
      return rv;

    } else {
      // We just looped so read sample
      return ReadNextFrame();
    }
  }

  frame = sample->frame;

  if (clipfadetype==FADEOUT && frame == (header.totalframes - FadeOutOffset - 32))
    Fade();
//	statusSAMPLE_READY=TRUE;
  if (trigger) /* Check trigger if set */
    CheckTriggers();

  return rv;
}

/********************************************************************\
						Init and Constructors
\********************************************************************/

Film::~Film()
{
  Close();
  Init();
}

Film::Film()
{
  sample = (PFrameSampleStr)staticSampleBuffer;
  ravout = NULL; 
  //shandle = NULL;
  trigger = 0;//&staticTriggers;
  clipfadetype = NOFADE;
  Init();
}

void Film::Init(void)
{
  statusPAUSE=FALSE;
  statusVIDEODIFFERENT=FALSE;
  statusPLAYING=FALSE;
  statusNO_AUDIO_TRACK=FALSE;
  statusSAMPLE_READY=FALSE;
  statusALL_WILL_FIT=FALSE;

  frame=0;
  loop=0;
  loopCount=0;

  //KILL_LINK(staticTriggers);

  //shandleid = -1;
  SetupRMCallback(NULL,NULL);
}

/********************************************************************\
Private Methods Here
\********************************************************************/

BOOL Film::Looping(void)
{
  loop--;
  if (loop < 1)
    return FALSE;

  loopCount++;
  reader.Loop();
  return TRUE;
}

BOOL Film::CheckSoundFormat(void)
{
  /* Only setup if we have too */
  WAVEFORMATEX wf;
  if (header.waveInfo.nSamplesPerSec==0)
  {	/* No audio track */
    statusNO_AUDIO_TRACK = TRUE;
    return TRUE; /* Don't change if 00000 */
  }
  statusNO_AUDIO_TRACK = FALSE;

  if (ravout->Sound()->GetFormat(&wf)==FALSE)
  {
    OutputDebugString("Error getting sound format.");
    return FALSE; /* Error in dsound */
  }

  wf.cbSize = 0;
  if (!memcmp(&wf,&header.waveInfo,sizeof(wf)))
    return TRUE; /* Already set correctly */

  /* Setup NEW audio format */
  return ravout->Sound()->Open(&header.waveInfo);
}

void Film::CheckTriggers(void)
{
  if (trigger)
  {
    for (int i=0; i<trigger->count(); i++)
    {
      PEventStr tr = trigger->ptrabs(i);
      if (frame==tr->frame)
      {
        (*(tr->fn))(this,tr->user); /* Callback User Function */
      }
    }
  }
}
